@@ -1,35 +1,53 @@
 use hudsucker::{
-    hyper::{Body, Response},
+    async_trait::async_trait,
+    hyper::{Body, Request, Response},
     rustls::internal::pemfile,
+    tungstenite::Message,
     *,
 };
 use log::*;
-use std::{future::Future, net::SocketAddr, pin::Pin};
+use std::net::SocketAddr;
 
 async fn shutdown_signal() {
     tokio::signal::ctrl_c()
         .await
-        .expect("failed to install CTRL+C signal handler");
+        .expect("Failed to install CTRL+C signal handler");
+}
+
+#[derive(Clone)]
+struct LogHandler {}
+
+#[async_trait]
+impl HttpHandler for LogHandler {
+    async fn handle_request(
+        &mut self,
+        _ctx: &HttpContext,
+        req: Request<Body>,
+    ) -> RequestOrResponse {
+        println!("{:?}", req);
+        RequestOrResponse::Request(req)
+    }
+
+    async fn handle_response(&mut self, _ctx: &HttpContext, res: Response<Body>) -> Response<Body> {
+        println!("{:?}", res);
+        res
+    }
+}
+
+#[derive(Clone)]
+struct NoopMessageHandler {}
+
+#[async_trait]
+impl MessageHandler for NoopMessageHandler {
+    async fn handle_message(&mut self, _ctx: &MessageContext, msg: Message) -> Option<Message> {
+        Some(msg)
+    }
 }
 
 #[tokio::main]
 async fn main() {
     env_logger::init();
 
-    let request_handler = |req| -> Pin<Box<dyn Future<Output = RequestOrResponse> + Send>> {
-        Box::pin(async {
-            println!("{:?}", req);
-            RequestOrResponse::Request(req)
-        })
-    };
-
-    let response_handler = |res| -> Pin<Box<dyn Future<Output = Response<Body>> + Send>> {
-        Box::pin(async {
-            println!("{:?}", res);
-            res
-        })
-    };
-
     let mut private_key_bytes: &[u8] = include_bytes!("ca/hudsucker.key");
     let mut ca_cert_bytes: &[u8] = include_bytes!("ca/hudsucker.pem");
     let private_key = pemfile::pkcs8_private_keys(&mut private_key_bytes)
@@ -45,10 +63,9 @@ async fn main() {
     let proxy_config = ProxyConfig {
         listen_addr: SocketAddr::from(([127, 0, 0, 1], 3000)),
         shutdown_signal: shutdown_signal(),
-        request_handler,
-        response_handler,
-        incoming_message_handler: |msg| Some(msg),
-        outgoing_message_handler: |msg| Some(msg),
+        http_handler: LogHandler {},
+        incoming_message_handler: NoopMessageHandler {},
+        outgoing_message_handler: NoopMessageHandler {},
         upstream_proxy: None,
         ca,
     };
