@@ -5,25 +5,29 @@
 //! - Modify websocket messages
 
 mod certificate_authority;
+mod decoder;
 mod error;
 mod proxy;
 mod rewind;
 
 use hyper::{
     client::HttpConnector,
+    server::conn::AddrStream,
     service::{make_service_fn, service_fn},
-    Body, Client, Request, Response, Server,
+    Body, Client, Request, Response, Server, Uri,
 };
 use hyper_proxy::{Proxy as UpstreamProxy, ProxyConnector};
 use hyper_rustls::HttpsConnector;
 use proxy::Proxy;
 use rustls::ClientConfig;
-use std::{convert::Infallible, future::Future, net::SocketAddr, pin::Pin};
+use std::{convert::Infallible, future::Future, net::SocketAddr};
 use tokio_tungstenite::tungstenite::Message;
 
 pub(crate) use rewind::Rewind;
 
+pub use async_trait;
 pub use certificate_authority::CertificateAuthority;
+pub use decoder::decode_response;
 pub use error::Error;
 pub use hyper;
 pub use hyper_proxy;
@@ -43,72 +47,68 @@ pub enum RequestOrResponse {
     Response(Response<Body>),
 }
 
-/// Handler for HTTP requests.
-///
-/// The handler will be called for each HTTP request. It can either return a modified request, or a
-/// response. If a request is returned, it will be sent to the upstream server. If a response is
-/// returned, it will be sent to the client.
-pub trait RequestHandler:
-    FnOnce(Request<Body>) -> Pin<Box<dyn Future<Output = RequestOrResponse> + Send>>
-    + Send
-    + Sync
-    + Clone
-    + 'static
-{
-}
-impl<T> RequestHandler for T where
-    T: FnOnce(Request<Body>) -> Pin<Box<dyn Future<Output = RequestOrResponse> + Send>>
-        + Send
-        + Sync
-        + Clone
-        + 'static
-{
+/// Context for HTTP requests and responses.
+#[derive(Clone, Debug, Eq, Hash, PartialEq)]
+pub struct HttpContext {
+    /// Address of the client that is sending the request.
+    pub client_addr: SocketAddr,
 }
 
-/// Handler for HTTP responses.
-///
-/// The handler will be called for each HTTP response. It can modify a response before it is
-/// forwarded to the client.
-pub trait ResponseHandler:
-    FnOnce(Response<Body>) -> Pin<Box<dyn Future<Output = Response<Body>> + Send>>
-    + Send
-    + Sync
-    + Clone
-    + 'static
-{
+/// Context for websocket messages.
+#[derive(Clone, Debug, Eq, Hash, PartialEq)]
+pub struct MessageContext {
+    /// Address of the client.
+    pub client_addr: SocketAddr,
+    /// URI of the server.
+    pub server_uri: Uri,
 }
-impl<T> ResponseHandler for T where
-    T: FnOnce(Response<Body>) -> Pin<Box<dyn Future<Output = Response<Body>> + Send>>
-        + Send
-        + Sync
-        + Clone
-        + 'static
-{
+
+/// Handler for HTTP requests and responses.
+///
+/// Each request/response pair is passed to the same instance of the handler.
+#[async_trait::async_trait]
+pub trait HttpHandler: Clone + Send + Sync + 'static {
+    /// The handler will be called for each HTTP request. It can either return a modified request,
+    /// or a response. If a request is returned, it will be sent to the upstream server. If a
+    /// response is returned, it will be sent to the client.
+    async fn handle_request(
+        &mut self,
+        context: &HttpContext,
+        request: Request<Body>,
+    ) -> RequestOrResponse;
+
+    /// The handler will be called for each HTTP response. It can modify a response before it is
+    /// forwarded to the client.
+    async fn handle_response(
+        &mut self,
+        context: &HttpContext,
+        response: Response<Body>,
+    ) -> Response<Body>;
 }
 
 /// Handler for websocket messages.
 ///
-/// The handler will be called for each websocket message. It can return an optional modified
-/// message. If None is returned the message will not be forwarded.
-pub trait MessageHandler:
-    FnMut(Message) -> Option<Message> + Send + Sync + Clone + 'static
-{
-}
-impl<T> MessageHandler for T where
-    T: FnMut(Message) -> Option<Message> + Send + Sync + Clone + 'static
-{
+/// Messages sent over the same websocket stream are passed to the same instance of the handler.
+#[async_trait::async_trait]
+pub trait MessageHandler: Clone + Send + Sync + 'static {
+    /// The handler will be called for each websocket message. It can return an optional modified
+    /// message. If None is returned the message will not be forwarded.
+    async fn handle_message(
+        &mut self,
+        context: &MessageContext,
+        message: Message,
+    ) -> Option<Message>;
 }
 
 /// Configuration for the proxy server.
 ///
 /// The proxy server can be configured with a number of options.
 #[derive(Clone)]
-pub struct ProxyConfig<F: Future<Output = ()>, R1, R2, W1, W2>
+pub struct ProxyConfig<F: Future<Output = ()>, H, M1, M2>
 where
-    R1: RequestHandler,
-    R2: ResponseHandler,
-    W1: MessageHandler,
-    W2: MessageHandler,
+    H: HttpHandler,
+    M1: MessageHandler,
+    M2: MessageHandler,
 {
     /// The address to listen on.
     pub listen_addr: SocketAddr,
@@ -116,14 +116,12 @@ where
     pub shutdown_signal: F,
     /// The certificate authority to use.
     pub ca: CertificateAuthority,
-    /// A handler for HTTP requests.
-    pub request_handler: R1,
-    /// A handler for HTTP responses.
-    pub response_handler: R2,
+    /// A handler for HTTP requests and responses.
+    pub http_handler: H,
     /// A handler for websocket messages sent from the client to the upstream server.
-    pub incoming_message_handler: W1,
+    pub incoming_message_handler: M1,
     /// A handler for websocket messages sent from the upstream server to the client.
-    pub outgoing_message_handler: W2,
+    pub outgoing_message_handler: M2,
     /// The upstream proxy to use.
     pub upstream_proxy: Option<UpstreamProxy>,
 }
@@ -131,43 +129,41 @@ where
 /// Attempts to start a proxy server using the provided configuration options.
 ///
 /// This will fail if the proxy server is unable to be started.
-pub async fn start_proxy<F, R1, R2, W1, W2>(
+pub async fn start_proxy<F, H, M1, M2>(
     ProxyConfig {
         listen_addr,
         shutdown_signal,
         ca,
-        request_handler,
-        response_handler,
+        http_handler,
         incoming_message_handler,
         outgoing_message_handler,
         upstream_proxy,
-    }: ProxyConfig<F, R1, R2, W1, W2>,
+    }: ProxyConfig<F, H, M1, M2>,
 ) -> Result<(), Error>
 where
     F: Future<Output = ()>,
-    R1: RequestHandler,
-    R2: ResponseHandler,
-    W1: MessageHandler,
-    W2: MessageHandler,
+    H: HttpHandler,
+    M1: MessageHandler,
+    M2: MessageHandler,
 {
     let client = gen_client(upstream_proxy);
 
-    let make_service = make_service_fn(move |_| {
+    let make_service = make_service_fn(move |conn: &AddrStream| {
         let client = client.clone();
         let ca = ca.clone();
-        let request_handler = request_handler.clone();
-        let response_handler = response_handler.clone();
+        let http_handler = http_handler.clone();
         let incoming_message_handler = incoming_message_handler.clone();
         let outgoing_message_handler = outgoing_message_handler.clone();
+        let client_addr = conn.remote_addr();
         async move {
             Ok::<_, Infallible>(service_fn(move |req| {
                 Proxy {
                     ca: ca.clone(),
                     client: client.clone(),
-                    request_handler: Some(request_handler.clone()),
-                    response_handler: response_handler.clone(),
+                    http_handler: http_handler.clone(),
                     incoming_message_handler: incoming_message_handler.clone(),
                     outgoing_message_handler: outgoing_message_handler.clone(),
+                    client_addr,
                 }
                 .proxy(req)
             }))
