@@ -1,40 +1,39 @@
 use crate::{
-    CertificateAuthority, MaybeProxyClient, MessageHandler, RequestHandler, RequestOrResponse,
-    ResponseHandler, Rewind,
+    CertificateAuthority, HttpContext, HttpHandler, MaybeProxyClient, MessageContext,
+    MessageHandler, RequestOrResponse, Rewind,
 };
-use futures::{sink::SinkExt, stream::StreamExt};
+use futures::{Sink, SinkExt, Stream, StreamExt};
 use http::uri::PathAndQuery;
 use hyper::{
     server::conn::Http, service::service_fn, upgrade::Upgraded, Body, Method, Request, Response,
+    Uri,
 };
 use log::*;
-use std::sync::Arc;
+use std::{net::SocketAddr, sync::Arc};
 use tokio::io::AsyncReadExt;
 use tokio_rustls::TlsAcceptor;
-use tokio_tungstenite::{connect_async, tungstenite, WebSocketStream};
+use tokio_tungstenite::{connect_async, tungstenite, tungstenite::Message, WebSocketStream};
 
 #[derive(Clone)]
-pub(crate) struct Proxy<R1, R2, W1, W2>
+pub(crate) struct Proxy<H, M1, M2>
 where
-    R1: RequestHandler,
-    R2: ResponseHandler,
-    W1: MessageHandler,
-    W2: MessageHandler,
+    H: HttpHandler,
+    M1: MessageHandler,
+    M2: MessageHandler,
 {
     pub ca: CertificateAuthority,
     pub client: MaybeProxyClient,
-    pub request_handler: Option<R1>,
-    pub response_handler: R2,
-    pub incoming_message_handler: W1,
-    pub outgoing_message_handler: W2,
+    pub http_handler: H,
+    pub incoming_message_handler: M1,
+    pub outgoing_message_handler: M2,
+    pub client_addr: SocketAddr,
 }
 
-impl<R1, R2, W1, W2> Proxy<R1, R2, W1, W2>
+impl<H, M1, M2> Proxy<H, M1, M2>
 where
-    R1: RequestHandler,
-    R2: ResponseHandler,
-    W1: MessageHandler,
-    W2: MessageHandler,
+    H: HttpHandler,
+    M1: MessageHandler,
+    M2: MessageHandler,
 {
     pub(crate) async fn proxy(self, req: Request<Body>) -> Result<Response<Body>, hyper::Error> {
         if req.method() == Method::CONNECT {
@@ -45,12 +44,11 @@ where
     }
 
     async fn process_request(mut self, req: Request<Body>) -> Result<Response<Body>, hyper::Error> {
-        let req = match (self
-            .request_handler
-            .take()
-            .expect("Request handler is None"))(req)
-        .await
-        {
+        let ctx = HttpContext {
+            client_addr: self.client_addr,
+        };
+
+        let req = match self.http_handler.handle_request(&ctx, req).await {
             RequestOrResponse::Request(req) => req,
             RequestOrResponse::Response(res) => return Ok(res),
         };
@@ -88,7 +86,7 @@ where
                 let server_socket = websocket.await.unwrap_or_else(|_| {
                     panic!("Failed to upgrade websocket connection for {}", uri)
                 });
-                self.handle_websocket(server_socket, &uri).await;
+                self.handle_websocket(server_socket, uri).await;
             });
 
             return Ok(res);
@@ -99,7 +97,7 @@ where
             MaybeProxyClient::Https(client) => client.request(req).await?,
         };
 
-        Ok((self.response_handler)(res).await)
+        Ok(self.http_handler.handle_response(&ctx, res).await)
     }
 
     async fn process_connect(self, req: Request<Body>) -> Result<Response<Body>, hyper::Error> {
@@ -148,59 +146,35 @@ where
         Ok(Response::new(Body::empty()))
     }
 
-    async fn handle_websocket(self, server_socket: WebSocketStream<Upgraded>, uri: &http::Uri) {
-        let (client_socket, _) = connect_async(uri)
+    async fn handle_websocket(self, server_socket: WebSocketStream<Upgraded>, uri: Uri) {
+        let (client_socket, _) = connect_async(&uri)
             .await
             .unwrap_or_else(|_| panic!("Failed to open websocket connection to {}", uri));
 
-        let (mut server_sink, mut server_stream) = server_socket.split();
-        let (mut client_sink, mut client_stream) = client_socket.split();
+        let (server_sink, server_stream) = server_socket.split();
+        let (client_sink, client_stream) = client_socket.split();
 
         let Proxy {
-            mut incoming_message_handler,
-            mut outgoing_message_handler,
+            incoming_message_handler,
+            outgoing_message_handler,
             ..
         } = self;
 
-        tokio::spawn(async move {
-            while let Some(message) = server_stream.next().await {
-                match message {
-                    Ok(message) => {
-                        let message = match incoming_message_handler(message) {
-                            Some(message) => message,
-                            None => continue,
-                        };
+        spawn_message_forwarder(
+            server_stream,
+            client_sink,
+            incoming_message_handler,
+            self.client_addr,
+            uri.clone(),
+        );
 
-                        match client_sink.send(message).await {
-                            Err(tungstenite::Error::ConnectionClosed) => (),
-                            Err(e) => error!("websocket send error: {}", e),
-                            _ => (),
-                        }
-                    }
-                    Err(e) => error!("websocket message error: {}", e),
-                }
-            }
-        });
-
-        tokio::spawn(async move {
-            while let Some(message) = client_stream.next().await {
-                match message {
-                    Ok(message) => {
-                        let message = match outgoing_message_handler(message) {
-                            Some(message) => message,
-                            None => continue,
-                        };
-
-                        match server_sink.send(message).await {
-                            Err(tungstenite::Error::ConnectionClosed) => (),
-                            Err(e) => error!("websocket send error: {}", e),
-                            _ => (),
-                        }
-                    }
-                    Err(e) => error!("websocket message error: {}", e),
-                }
-            }
-        });
+        spawn_message_forwarder(
+            client_stream,
+            server_sink,
+            outgoing_message_handler,
+            self.client_addr,
+            uri,
+        );
     }
 
     async fn serve_websocket(self, stream: Rewind<Upgraded>) -> Result<(), hyper::Error> {
@@ -274,3 +248,36 @@ where
             .await
     }
 }
+
+fn spawn_message_forwarder(
+    mut stream: impl Stream<Item = Result<Message, tungstenite::Error>> + Unpin + Send + 'static,
+    mut sink: impl Sink<Message, Error = tungstenite::Error> + Unpin + Send + 'static,
+    mut handler: impl MessageHandler,
+    client_addr: SocketAddr,
+    uri: Uri,
+) {
+    let ctx = MessageContext {
+        client_addr,
+        server_uri: uri,
+    };
+
+    tokio::spawn(async move {
+        while let Some(message) = stream.next().await {
+            match message {
+                Ok(message) => {
+                    let message = match handler.handle_message(&ctx, message).await {
+                        Some(message) => message,
+                        None => continue,
+                    };
+
+                    match sink.send(message).await {
+                        Err(tungstenite::Error::ConnectionClosed) => (),
+                        Err(e) => error!("websocket send error: {}", e),
+                        _ => (),
+                    }
+                }
+                Err(e) => error!("websocket message error: {}", e),
+            }
+        }
+    });
+}
